"use strict";
/**
 * Created by bert on 07.07.16.
 */
Object.defineProperty(exports, "__esModule", { value: true });
const reob_1 = require("./reob");
const wm = require("web-methods");
const eventemitter = require("eventemitter2");
const Annotations_1 = require("./Annotations");
/**
 * This is class represents the startingpoint from the website. Use it to register services and load objects.
 */
class Client {
    constructor(host, port) {
        this.singletons = {};
        /**
         * @hidden
         */
        this.webMethodRunning = false;
        // if( !omm.isInitialized() )
        //     throw new Error("omm is not initialized.");
        var endpointUrl = "http://" + host + ":" + port + "/methods";
        this.webMethods = new wm.WebMethods(endpointUrl);
        this.serializer = new reob_1.Serializer();
        this.eventEmitter = new eventemitter.EventEmitter2();
    }
    /**
     * Add a service to the client. After this you can call functions decorated with @remote so
     */
    addService(serviceName, service) {
        // var serviceName = omm.getServiceName( omm.getClass(service) );
        // if( !serviceName ){
        //     console.log(service, omm.getClass(service));
        //     throw new Error("Registered service has no @Service(<name>) decorator." );
        // }
        this.singletons[serviceName] = service;
        var serviceClass = Annotations_1.getClass(service);
        Annotations_1.getRemoteFunctionNames(serviceClass).forEach((remoteFunctionName) => {
            var options = Annotations_1.getMethodOptions(serviceClass, remoteFunctionName);
            if (!options.name) {
                options.name = serviceName + "." + options.propertyName;
            }
        });
        reob_1.SerializationPath.setObjectContext(service, undefined, this, undefined);
    }
    /**
     * Load an object with a given id from a collection.
     */
    load(collectionName, id) {
        return this.loadDocument(collectionName, id).then((doc) => {
            var o = this.serializer.toObject(doc, this);
            return o;
        });
    }
    /**
     * Load a document with a given id from a collection.
     */
    loadDocument(collectionName, id) {
        return this.webMethods.call("get", collectionName, id, this.userData).then((document) => {
            return document;
        });
    }
    /**
     * @hidden
     * @param methodName
     * @param objectId
     * @param args
     * @returns {Promise<T>|Promise<T|U>}
     */
    call(methodName, objectId, args) {
        var webArgs = [];
        // convert the arguments to their origin
        for (var i in args) {
            if (args[i]) {
                if (args[i]._serializationPath) {
                    webArgs[i] = args[i]._serializationPath.toString();
                }
                else if (typeof args[i] != "function") {
                    webArgs[i] = this.serializer.toDocument(args[i]);
                }
                else {
                    throw new Error("There can not be parameters of type 'function' in the arguments when calling a web method with name " + methodName);
                }
            }
            else {
                webArgs[i] = args[i];
            }
        }
        // prepend
        webArgs.unshift(this.userData);
        webArgs.unshift(objectId);
        webArgs.unshift(methodName);
        var p = this.webMethods.call.apply(this.webMethods, webArgs);
        return p.then((result) => {
            // convert the result from json to an object
            var obje;
            if (result) {
                obje = this.serializer.toObject(result.document, this);
            }
            return obje;
        }).catch((reason) => {
            if (reason.message == "XHR error") {
                var e = new Error("Network error");
                this.emitNetworkError(e);
                throw e;
            }
            else
                throw reason;
        });
    }
    /**
     * Register listeners that are called whenever a network error happens.
     */
    onNetworkError(f) {
        this.eventEmitter.on("network-error", f);
    }
    /**
     * Remove listeners that are called whenever a network error happens.
     */
    removeNetworkErrorListener(f) {
        this.eventEmitter.removeListener("network-error", f);
    }
    /**
     * @hidden
     */
    emitNetworkError(error) {
        this.eventEmitter.emit("network-error", error);
    }
    /**
     * @hidden
     */
    getSingletonKey(o) {
        for (var i in this.singletons) {
            if (this.singletons[i] == o)
                return i;
        }
        return undefined;
    }
    /**
     * @hidden
     */
    webMethod(entityClass, functionName, object, originalFunction, args) {
        if (this.webMethodRunning) {
            return originalFunction.apply(object, args);
        }
        var objectContext = reob_1.SerializationPath.getObjectContext(object);
        var sp = objectContext.serializationPath ? objectContext.serializationPath : undefined;
        var serviceKey = this.getSingletonKey(object);
        var key = serviceKey || (sp ? sp.toString() : undefined);
        var r;
        var options = Annotations_1.getMethodOptions(entityClass, functionName);
        if (key) {
            var name = options.name;
            if (!name) {
                if (serviceKey)
                    name = serviceKey + "." + functionName;
                else
                    name = Annotations_1.getClassName(entityClass) + "." + functionName;
            }
            r = this.call(name, key, args);
        }
        if (!options.serverOnly || !key) {
            var rOriginal;
            this.webMethodRunning = true;
            try {
                rOriginal = Promise.resolve(originalFunction.apply(object, args));
            }
            catch (error) {
                rOriginal = Promise.reject(error);
            }
            finally {
                this.webMethodRunning = false;
            }
            if (!r)
                r = rOriginal;
            else {
                rOriginal.catch(() => {
                    // hide exception, as the result is coming from the server
                });
            }
            reob_1.SerializationPath.updateObjectContexts(object, this, undefined);
        }
        return r;
    }
    /**
     * Use this method to set any data that you'd like to use for authentication and authorization purposes on the server.
     * It is transmitted with every request.
     */
    setUserData(ud) {
        this.userData = ud;
    }
}
exports.Client = Client;
//# sourceMappingURL=Client.js.map