"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const reob = require("./serverModule");
const uuid = require("uuid");
/**
 * @hidden
 */
function getDefaultCollectionName(t) {
    return reob.getClassName(t);
}
/**
 * @hidden
 */
class EventNames {
}
EventNames.onBeforeRemove = "onBeforeRemove";
EventNames.onAfterRemove = "onAfterRemove"; // only id
EventNames.onBeforeInsert = "onBeforeInsert";
EventNames.onAfterInsert = "onAfterInsert";
EventNames.onBeforeUpdate = "onBeforeUpdate";
EventNames.onAfterUpdate = "onAfterUpdate";
EventNames.onBeforeSave = "onBeforeSave";
EventNames.onDuringUpdate = "onDuringUpdate";
class Collection {
    constructor(entityClass, collectionName) {
        this.eventListenerRegistry = new reob.EventListenerRegistry();
        // the handler function for the collection updates of objects loaded via this collection
        this.updating = false;
        this.serializer = new reob.Serializer();
        if (!collectionName)
            collectionName = getDefaultCollectionName(entityClass);
        this.name = collectionName;
        this.theClass = entityClass;
    }
    removeAllListeners() {
        this.eventListenerRegistry.removeAllListeners();
    }
    /**
     * Adds an event listener that is called before an object is removed from the collection. The listener can cancel the operation.
     * @param f the listener
     */
    onBeforeRemove(f) {
        this.eventListenerRegistry.on(EventNames.onBeforeRemove, undefined, f);
    }
    /**
     * Adds an event listener that is called after an object is removed from the collection.
     * @param f the listener
     */
    onAfterRemove(f) {
        this.eventListenerRegistry.on(EventNames.onAfterRemove, undefined, f);
    }
    /**
     * Adds an event listener that is called after an object is inserted into the collection. The listener can cancel the operation.
     * @param f the listener
     */
    onBeforeInsert(f) {
        this.eventListenerRegistry.on(EventNames.onBeforeInsert, undefined, f);
    }
    /**
     * Adds an event listener that is called after an object is inserted into the collection.
     * @param f the listener
     */
    onAfterInsert(f) {
        this.eventListenerRegistry.on(EventNames.onAfterInsert, undefined, f);
    }
    /**
     * Adds an event listener that is called before a collection is updated with an updated object.
     * @param f the listener
     */
    onBeforeSave(f) {
        this.eventListenerRegistry.on(EventNames.onBeforeSave, undefined, f);
    }
    /**
     * Adds an event listener that is called before a remote function updates an object. The listener can cancel the operation.
     * @param f the listener
     */
    onBeforeUpdate(cls, functionName, f) {
        this.registerUpdateEventListener(EventNames.onBeforeUpdate, cls, functionName, f);
    }
    /**
     * Adds an event listener that is called after a remote function updates an object.
     * @param f the listener
     */
    onAfterUpdate(cls, functionName, f) {
        this.registerUpdateEventListener(EventNames.onAfterUpdate, cls, functionName, f);
    }
    registerUpdateEventListener(topic, cls, functionName, f) {
        var subTopic = undefined;
        if (!f && typeof functionName === "undefined" && typeof cls === "function") {
            f = cls;
        }
        else if (typeof cls !== "undefined" && typeof functionName !== "undefined" && f && reob.getClassName(cls)) {
            if (reob.getCollectionUpdateFunctionNames(cls).indexOf(functionName) == -1)
                throw new Error("'" + functionName + "' is not a collection updating function on the class/entity " + reob.getClassName(cls) + ".");
            subTopic = reob.getClassName(cls) + "." + functionName;
        }
        else {
            throw new Error("Illegal combination of parameters.");
        }
        this.eventListenerRegistry.on(topic, subTopic, f);
    }
    /**
     * Adds an event listener that is called after an update is completed in case events were emitted using reob.emit(topic).
     * @param topic optional the event name that's emitted by reob.emit() in the update function
     * @param f the listener
     */
    onDuringUpdate(topic, f) {
        if (typeof topic == "function") {
            f = topic;
            topic = undefined;
        }
        this.eventListenerRegistry.on(EventNames.onDuringUpdate, topic, f);
    }
    emit(t, subTopic, ...eventArgs) {
        return this.eventListenerRegistry.emit(t, subTopic, eventArgs);
    }
    setMongoCollection(db) {
        this.mongoCollection = db.collection(this.name);
    }
    /**
     * Gets the name of the collection.
     * @returns {string}
     */
    getName() {
        return this.name;
    }
    /**
     * Returns the underlying mongo collection.
     * @returns {any}
     */
    getMongoCollection() {
        if (!this.mongoCollection) {
            throw new Error("No Mongo Collection set. Use 'setMongoCollection' first.");
        }
        return this.mongoCollection;
    }
    /**
     * Loads an object from the collection by its id.
     * @param id {string} the id
     * @returns {T} the object or undefined if it wasn't found
     */
    getById(id, request) {
        return this.find({
            "_id": id
        }, request).then((values) => {
            if (values.length)
                return values[0];
            else
                return undefined;
        });
    }
    /**
     * Finds objects based on a selector.
     * @param {object} findCriteria the mongo selector
     * @returns {Array<T>}
     * @protected
     */
    find(findCriteria, request) {
        return this.cursorToObjects(this.getMongoCollection().find(findCriteria), request);
    }
    cursorToObjects(c, request) {
        var cursor = c;
        return cursor.toArray().then((documents) => {
            var objects = [];
            for (var i = 0; i < documents.length; i++) {
                var document = documents[i];
                objects[i] = this.serializer.toObject(document, this, this.getEntityClass(), new reob.SerializationPath(this.getName(), document._id), request);
            }
            return objects;
        });
    }
    /**
     * Gets all objects in a collection.
     * @returns {Array<T>}
     */
    getAll(request) {
        return this.find({}, request);
    }
    getByIdOrFail(id, request) {
        return this.getById(id, request).then((t) => {
            if (!t)
                return Promise.reject(new Error("Not found"));
            else
                return Promise.resolve(t);
        });
    }
    /**
     * Removes an entry from a collection
     * @param id {string} the id of the object to be removed from the collection
     * @callback cb the callback that's called once the object is removed or an error happend
     */
    remove(id, request) {
        if (!id)
            return Promise.reject(new Error("Trying to remove an object without an id."));
        return this.emit(EventNames.onBeforeRemove, undefined, id, request).then(() => {
            return this.getMongoCollection().remove({ _id: id }).then((result) => {
                return this.emit(EventNames.onAfterRemove, undefined, id, request).then(() => true);
            });
        });
    }
    getByIds(ids, request) {
        return this.find({
            _id: { $in: ids }
        }, request).then((objs) => {
            var result = {};
            objs.forEach((o) => {
                result[reob.getId(o)] = o;
            });
            return result;
        });
    }
    /**
     * Performs an update on an object in the collection. After the update the object is attempted to be saved to
     * the collection. If the object has changed between the time it was loaded and the time it is saved, the whole
     * process is repeated. This means that the updateFunction might be called more than once.
     * @param id - the id of the object
     * @param updateFunction - the function that alters the loaded object
     */
    update(sp, updateFunction, request) {
        if (!sp || !updateFunction)
            return Promise.reject(new Error("update function or serialiationPath parameter missing"));
        return this.updateOnce(sp, updateFunction, 0, request, undefined, false, undefined).then((cru) => {
            return cru.result;
        });
    }
    updateOnce(sp, updateFunction, attempt, request, functionName, didSendBeforeUpdateEvents, originalArguments) {
        var documentPromise = this.getMongoCollection().find({
            _id: sp.getId()
        }).toArray().then((documents) => {
            var document = documents[0];
            if (!document) {
                return Promise.reject(new Error("No document found for id: " + sp.getId()));
            }
            return document;
        });
        var currentSerialPromise = documentPromise.then((doc) => {
            return doc.serial;
        });
        var rootObjectPromise = documentPromise.then((doc) => {
            // do not user a handler. If the original function is called, there shouldn't be any more collection updates
            // happening
            return this.serializer.toObject(doc, undefined, this.getEntityClass());
        });
        var objectPromise = rootObjectPromise.then((rootObject) => {
            return sp.getSubObject(rootObject);
        });
        var beforeUpdateEventSentPromise = Promise.all([rootObjectPromise, objectPromise, documentPromise]).then((values) => {
            var updateEvent = {
                rootObject: values[0],
                subObject: values[1],
                functionName: functionName,
                args: originalArguments,
                result: undefined,
                request: request,
                beforeUpdateDocument: values[2],
                afterUpdateDocument: undefined,
                data: undefined
            };
            return this.emit(EventNames.onBeforeUpdate, undefined, updateEvent).then(() => { });
        });
        var resultPromise = Promise.all([objectPromise, rootObjectPromise, beforeUpdateEventSentPromise]).then((values) => {
            var object = values[0];
            var rootObject = values[1];
            // call the update function
            var result = {};
            result.events = [];
            reob.setEmitHandler((topic, data) => {
                result.events.push({ topic: EventNames.onDuringUpdate, subTopic: topic, data: data });
            });
            // This is where the update function is called
            result.result = updateFunction(object);
            result.object = object;
            // maybe there are new objects in the tree that starts with root object. those need the context.
            reob.SerializationPath.setObjectContext(rootObject, new reob.SerializationPath(this.getName(), reob.getId(rootObject)), this, request);
            reob.SerializationPath.updateObjectContexts(rootObject, this, request);
            return result;
        });
        var updatePromise = Promise.all([objectPromise, currentSerialPromise, resultPromise, rootObjectPromise, documentPromise]).then((values) => {
            var currentSerial = values[1];
            var rootObject = values[3];
            var documentToSave = this.serializer.toDocument(rootObject);
            return this.emit(EventNames.onBeforeSave, undefined, rootObject, request).then(() => {
                documentToSave.serial = (currentSerial || 0) + 1;
                return this.getMongoCollection().replaceOne({
                    _id: reob.getId(rootObject),
                    serial: currentSerial
                }, documentToSave).then((updateResult) => {
                    updateResult['documentToSave'] = documentToSave;
                    return updateResult;
                });
            });
        });
        return Promise.all([resultPromise, updatePromise, rootObjectPromise, documentPromise]).then((values) => __awaiter(this, void 0, void 0, function* () {
            var result = values[0];
            var updateResult = values[1];
            var rootObject = values[2];
            var documentPre = values[3];
            // verify that that went well
            if (updateResult.modifiedCount == 1) {
                var cr = {
                    events: result.events,
                    rootObject: rootObject,
                    object: result.object,
                    result: result.result,
                    rootDocumentPre: documentPre,
                    rootDocumentPost: updateResult.documentToSave
                };
                return cr;
            }
            else if (updateResult.modifiedCount > 1) {
                return Promise.reject(new Error("Reob updated more than 1 document. This is not expected."));
            }
            else if (attempt < 10) {
                return yield this.updateOnce(sp, updateFunction, attempt + 1, request, functionName, true, originalArguments);
            }
            else {
                return Promise.reject(new Error("Reob tried 10 times to update the document. This happens if there is a lot of concurrency."));
            }
        })).then((r) => {
            // send  'duringUpdate' events
            var promises = [];
            r.events.forEach((t) => {
                var updateEvent = {
                    rootObject: r.rootObject,
                    subObject: r.object,
                    functionName: functionName,
                    args: originalArguments,
                    result: r.result,
                    request: request,
                    beforeUpdateDocument: r.rootDocumentPre,
                    afterUpdateDocument: r.rootDocumentPost,
                    data: t.data
                };
                promises.push(Promise.resolve(this.emit(t.topic, t.subTopic, updateEvent)));
            });
            // we ignore rejected promises from during and after events, we just want them to happen before this finishes.
            var duringSentPromise = Promise.all(promises).catch(() => { }).then(() => {
            });
            // send  'afterUpdate' events
            var afterSentPromise = duringSentPromise.then(() => {
                var updateEvent = {
                    rootObject: r.rootObject,
                    subObject: r.object,
                    functionName: functionName,
                    args: originalArguments,
                    result: r.result,
                    request: request,
                    beforeUpdateDocument: r.rootDocumentPre,
                    afterUpdateDocument: r.rootDocumentPost,
                    data: undefined
                };
                return this.emit(EventNames.onAfterUpdate, undefined, updateEvent);
            }).catch(() => { });
            return afterSentPromise.then(() => r);
        });
    }
    /**
     * Inserts an object into the collection
     * @param p the object
     * @param {reob.Collection~insertCallback} callback
     * @returns {string} the id of the new object
     */
    insert(p, request) {
        // TODO make sure that this is unique
        var idPropertyName = reob.getIdPropertyName(this.theClass);
        var id = p[idPropertyName];
        if (!id) {
            p[idPropertyName] = uuid.v1(); //new mongodb.ObjectID().toString();
            id = p[idPropertyName];
        }
        return this.emit(EventNames.onBeforeInsert, undefined, p, request).then(() => {
            //if( omm.verbose )console.log("insert not cancelled");
            var doc = this.serializer.toDocument(p);
            doc.serial = 0;
            //if( omm.verbose )console.log( "inserting document: ", doc);
            return this.getMongoCollection().insertOne(doc).then(() => {
                reob.SerializationPath.setObjectContext(p, new reob.SerializationPath(this.getName(), id), this, request);
                reob.SerializationPath.updateObjectContexts(p, this, request);
                return this.emit(EventNames.onAfterInsert, undefined, p, request).then(() => id);
            });
        });
    }
    getEntityClass() {
        return this.theClass;
    }
    collectionUpdate(entityClass, functionName, object, originalFunction, args, request) {
        if (this.updating) {
            return originalFunction.apply(object, args);
        }
        // var rootObject;
        // var objectPromise:Promise<any>;
        //
        // var rootObjectPromise:Promise<any>;
        //
        //
        // rootObjectPromise =  this.getById(sp.getId(), request);
        // objectPromise = rootObjectPromise.then((rootObject:any)=>{
        //     return sp.getSubObject(rootObject);
        // });
        var objectContext = reob.SerializationPath.getObjectContext(object);
        var sp = objectContext.serializationPath;
        return this.updateOnce(sp, (subObject) => {
            this.updating = true;
            try {
                var r2 = originalFunction.apply(subObject, args);
                return r2;
            }
            finally {
                this.updating = false;
            }
        }, 0, request, functionName, false, args).then((cru) => {
            return cru.result;
        });
        // return Promise.all([objectPromise,rootObjectPromise]).then((values:any[])=>{
        // });
    }
}
exports.Collection = Collection;
//# sourceMappingURL=Collection.js.map