"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
/**
 * @hidden
 */
class EventListenerRegistry {
    constructor() {
        this.globalListeners = [];
        this.topicListeners = {};
        this.subTopicListeners = {};
    }
    on(topic, subTopic, handler) {
        if (typeof topic === "undefined" && typeof subTopic === "undefined" && typeof handler === "undefined") {
            this.addGlobalListener(handler);
        }
        else if (topic && typeof subTopic === "undefined" && handler) {
            this.addTopicListener(topic, handler);
        }
        else if (topic && subTopic && handler) {
            this.addSubTopicListener(topic, subTopic, handler);
        }
        else {
            throw new Error("Illegal combination of parameters for EventListenerRegistry.on .");
        }
    }
    removeAllListeners() {
        this.globalListeners = [];
        this.topicListeners = {};
        this.subTopicListeners = {};
    }
    /**
     * Takes a function and converts its result to a promise. If the function throws an exception a rejected Promise is
     * returned.
     * @param f - the function to convert
     * @returns {Promise}
     */
    static castToPromise(f) {
        return new Promise((resolve, reject) => {
            try {
                var r = f();
                resolve(r);
            }
            catch (e) {
                reject(e);
            }
        });
    }
    emit(topic, subTopic, eventArgs) {
        var promises = [];
        if (typeof topic !== "undefined" && typeof subTopic !== "undefined" && this.subTopicListeners[topic] && this.subTopicListeners[topic][subTopic]) {
            this.subTopicListeners[topic][subTopic].forEach((h) => {
                promises.push(EventListenerRegistry.castToPromise(() => {
                    return h.apply(undefined, eventArgs);
                }));
            });
        }
        if (typeof topic !== "undefined" && this.topicListeners[topic]) {
            this.topicListeners[topic].forEach((h) => {
                promises.push(EventListenerRegistry.castToPromise(() => {
                    return h.apply(undefined, eventArgs);
                }));
            });
        }
        this.globalListeners.forEach((h) => {
            promises.push(EventListenerRegistry.castToPromise(() => {
                return h.apply(undefined, eventArgs);
            }));
        });
        return Promise.all(promises).then(() => { });
    }
    removeEventListener(topic, subTopic, handler) {
        if (typeof topic == "function" && !subTopic && !handler) {
            this.removeGlobalListener(handler);
        }
        else if (typeof topic === "string" && typeof subTopic === "function" && typeof handler === "undefined") {
            this.removeTopicListener(topic, handler);
        }
        else if (typeof topic === "string" && typeof subTopic === "function" && typeof handler === "function") {
            this.removeSubTopicListener(topic, subTopic, handler);
        }
        else {
            throw new Error("Illegal combination of parameters for EventListenerRegistry.removeEventListener .");
        }
    }
    addGlobalListener(handler) {
        this.globalListeners.push(handler);
    }
    removeGlobalListener(handler) {
        var index = this.globalListeners.indexOf(handler);
        if (index != -1) {
            this.globalListeners.splice(index, 1);
        }
    }
    addTopicListener(topic, handler) {
        if (!this.topicListeners[topic]) {
            this.topicListeners[topic] = [];
        }
        this.topicListeners[topic].push(handler);
    }
    removeTopicListener(topic, handler) {
        if (this.topicListeners[topic]) {
            var index = this.topicListeners[topic].indexOf(handler);
            if (index != -1) {
                this.topicListeners[topic].splice(index, 1);
            }
        }
    }
    addSubTopicListener(topic, subTopic, handler) {
        if (!this.subTopicListeners[topic]) {
            this.subTopicListeners[topic] = {};
        }
        if (!this.subTopicListeners[topic][subTopic]) {
            this.subTopicListeners[topic][subTopic] = [];
        }
        this.subTopicListeners[topic][subTopic].push(handler);
    }
    removeSubTopicListener(topic, subTopic, handler) {
        if (this.subTopicListeners[topic] && this.subTopicListeners[topic][subTopic]) {
            var index = this.subTopicListeners[topic][subTopic].indexOf(handler);
            if (index != -1) {
                this.subTopicListeners[topic][subTopic].splice(index, 1);
            }
        }
    }
}
exports.EventListenerRegistry = EventListenerRegistry;
//# sourceMappingURL=EventListenerRegistry.js.map